/*************************************************************************\
*                  LCD - Liquid Crystal Display (8.bit mode)              *
\*************************************************************************/

#include "lcd.h"                     // header file for LCD SC1602A

#if defined (__C51__)
#include <ABSACC.H>
#include <intrins.h>                 // intrinsic functions for C51
#define NOP _nop_()
#define   LCD_DATA_WR	XBYTE[0x7FFF]
#define   LCD_CMD_WR	XBYTE[0x3FFF]
#define   LCD_DATA_RD	XBYTE[0xFFFF]
#define   LCD_CMD_RD	XBYTE[0xBFFF]
#else
 #ifdef SDCC
#define NOP _asm nop _endasm      	// definition of a single nop command
xdata at 0x7FFF unsigned char LCD_DATA_WR;
xdata at 0xFFFF unsigned char LCD_DATA_RD;
xdata at 0x3FFF unsigned char LCD_CMD_WR;
xdata at 0xBFFF unsigned char LCD_CMD_RD;
 #endif
#endif

#ifdef FOUR_BITS
#define LCD_EN_LOW  {LCD_EN = 0; NOP;}
#define LCD_EN_HIGH {LCD_EN = 1; NOP;}

#define P1_3_STATE        0x08       // 0x08 ... input pin P1.3
									 // 0x00 ... output pin P1.3
						   //0123456789012345678901234567890123456789
code char txt_hex_str[]   = "0123456789!\x022#$%&'()*+,-./:;<=>?@{|}~\x7F";
code char txt_large_str[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijkmno";
code char txt_msg0_str[]  = "www.promislbc.cz";
code char txt_msg1_str[]  = "LCD display     ";
code char txt_msg2_str[]  = "simulator based ";
code char txt_msg3_str[]  = "on HD44780 chip ";
#else
code char txt_hex_str[]   = "0123456789ABCDEFGHIJ";
code char txt_large_str[] = "KLMNOPQRSTUVWXYZabcd";
code char txt_small_str[] = "efghijklmnopqrstuvwx";
code char txt_sym_str[]	  = "yz!\x022#$%&'()*+,-./:=?";
code char txt_msg0_str[]  = "-=www.promislbc.cz=-";
code char txt_msg1_str[]  = "presents LCD display";
code char txt_msg2_str[]  = "simulator of Hitachi";
code char txt_msg3_str[]  = "HD44780 driver     ";
#endif

//******************************************************************************
// --- move LCD cursor to X,Y position --- //
void LCD_Cursor_YX( unsigned char row, unsigned char column ) {
unsigned char i;
#ifdef CHARS_PER_LINE_16
  switch( row ) {
	default:
	case 0 : i = 0x00; break;
	case 1 : i = 0x40; break;
	case 2 : i = 16; break;
	case 3 : i = 16+0x40; break;
  }
#else
  switch( row ) {
	default:
	case 0 : i = 0x00; break;
	case 1 : i = 0x40; break;
	case 2 : i = 20; break;
	case 3 : i = 20+0x40; break;
  }
#endif
  LCD_SendCmd(0x80 | (i+column));
}

// --- send character to LCD --- //
void LCD_Putch( unsigned char val ) {
  LCD_SendData( val );
}

// --- display string on LCD --- //
void LCD_Puts( unsigned char *str ) {
  while(*str) LCD_Putch( (unsigned char) *str++);
}

void DelayUs(unsigned char i) {
  while (--i);
}

void DelayMs(unsigned int i) {
  i *= 4;
  while (i--) {
	DelayUs(250);
	if (!ButtonR) return;
	if (!ButtonL) return;
  }
}

// --- send data to LCD --- //
void LCD_SendVal( unsigned char val, bit RS ) {
#ifdef FOUR_BITS
  LCD_RS = RS; // RS ... 0 (command), RS ... 1 (data)
  LCD_EN_HIGH;
  LCD_PORT = (LCD_PORT & 0x0F) | (val & 0xF0) | P1_3_STATE;  // high nibble
  LCD_EN_LOW;

  LCD_EN_HIGH;
  LCD_PORT = (LCD_PORT & 0x0F) | (val << 4 )| P1_3_STATE;    // low nibble
  LCD_EN_LOW;
#else
  if (RS) LCD_DATA_WR=val;	//data
  else LCD_CMD_WR=val;		//command
#endif
  LCD_Wait(); // The busy flag must be checked after the 4-bit data
			  // has been transferred twice.
}

unsigned char LCD_Read(bit RData) {
unsigned char cont;
#ifdef FOUR_BITS
	LCD_RS   = RData;
	LCD_RW   = 1;
	LCD_EN   = 0;
	LCD_PORT = LCD_PORT | 0xF0 | P1_3_STATE;      // RS = 0, RW = 1, [DB4..DB7] ... 1
	LCD_EN = 1;
	cont = LCD_PORT & 0xF0;
	LCD_EN = 0;
	LCD_EN = 1;
	cont = cont | ((LCD_PORT & 0xF0) >> 4);
	LCD_EN = 0;
	LCD_RW = 0;
#else
  if (RData) cont=LCD_DATA_RD;
  else cont=LCD_CMD_RD;
#endif
	return cont;
}

// --- wait until LCD finish current operation --- //
void LCD_Wait( void ) {
	while (LCD_Read(0) & 0x80);			// check busy flag
}

// --- Initializing by Instruction initialization --- //
void LCD_Init(void) {					// all delays are vital
#define INIT_VAL	0x30
  DelayMs(100);        // power on delay - wait more than 15 ms
#ifdef FOUR_BITS
  LCD_RS   = 0;
  LCD_RW   = 0;
  LCD_EN_HIGH;
  LCD_PORT = (LCD_PORT & 0x0F) | INIT_VAL | P1_3_STATE;
  LCD_EN_LOW;
  DelayMs(5);          // wait more than 4.1 ms

  LCD_EN_HIGH;
  LCD_PORT = (LCD_PORT & 0x0F) | INIT_VAL | P1_3_STATE;
  LCD_EN_LOW;          //	---- second line disapears ----
  DelayUs(100);        // wait more than 100us

  LCD_EN_HIGH;
  LCD_PORT = (LCD_PORT & 0x0F) | INIT_VAL | P1_3_STATE;
  LCD_EN_LOW;          // ---- second line appears back but damaged ----
  DelayUs(100);        // wait more than 100us

  LCD_EN_HIGH;
  LCD_PORT = (LCD_PORT & 0x0F) | 0x20 | P1_3_STATE;  // set 4 bit mode
  LCD_EN_LOW;           // ----- damages first line ----
  DelayUs(100);        // wait more than 100us
  LCD_SendCmd(0x28);   // 4 bit mode, 1/16 duty, 5x8 font
#else
  LCD_CMD_WR=INIT_VAL;
  DelayMs(5);          // wait more than 4.1 ms
  LCD_CMD_WR=INIT_VAL;
  DelayUs(100);        // wait more than 100us
  LCD_CMD_WR=INIT_VAL;
  DelayUs(100);        // wait more than 100us
  LCD_SendCmd(0x38);   // 8 bit mode, 1/16 duty, 5x8 font
#endif
  LCD_SendCmd(0x08);   // display off
  LCD_SendCmd(0x01);   // display clear
  LCD_SendCmd(0x06);   // entry mode
  LCD_SendCmd(0x0C);   // display on, blinking cursor off
}

void main(void) {
  LCD_Init();                     // LCD initialization by instruction - 4-bit mode

  LCD_SendCmd(0x0F);
  while(1) {
	LCD_SendCmd(1);
#ifdef FOUR_BITS
	LCD_Puts(txt_msg0_str);
	LCD_Cursor_YX( 1, 0 );
	LCD_Puts(txt_msg1_str);
	DelayMs(2000);
	if (!ButtonR) break;
	if (!ButtonL) break;

	LCD_SendCmd(1);
	LCD_Puts(txt_msg2_str);
	LCD_Cursor_YX( 1, 0 );
	LCD_Puts(txt_msg3_str);
	DelayMs(2000);
	if (!ButtonR) break;
	if (!ButtonL) break;

	LCD_SendCmd(1);
	LCD_Puts(txt_hex_str);
	LCD_Cursor_YX( 1, 0 );
	LCD_Puts(txt_large_str);
	DelayMs(2000);
	if (!ButtonR) break;
	if (!ButtonL) break;
#else
	LCD_Puts(txt_msg0_str);
	LCD_Cursor_YX( 1, 0 );
	LCD_Puts(txt_msg1_str);
	LCD_Cursor_YX( 2, 0 );
	LCD_Puts(txt_msg2_str);
	LCD_Cursor_YX( 3, 0 );
	LCD_Puts(txt_msg3_str);
	DelayMs(2000);
	if (!ButtonR) break;
	if (!ButtonL) break;

	LCD_SendCmd(1);
	LCD_Puts(txt_hex_str);
	LCD_Cursor_YX( 1, 0 );
	LCD_Puts(txt_large_str);
	LCD_Cursor_YX( 2, 0 );
	LCD_Puts(txt_small_str);
	LCD_Cursor_YX( 3, 0 );
	LCD_Puts(txt_sym_str);
#endif
	DelayMs(2000);
	if (!ButtonR) break;
	if (!ButtonL) break;
  }
  while (1) {
	if (!ButtonR) {
		LCD_SendCmd(0x1C);
		while (!ButtonR);
	}
	else if (!ButtonL) {
		LCD_SendCmd(0x18);
		while (!ButtonL);
	}
	while (ButtonL & ButtonR);
  }
}


